#include "font.h"
#include "mip.h"

#ifdef DRAW_OPENGL
#include "image.h"
#include <GL/glew.h>

#include <cstring>

Font::Font()
{
	img = 0;
}
Font::Font(Image *img_)
{
	img = 0;
	load(img_);
}
Font::~Font()
{
}

void Font::load(Image *img_)
{
	spacing = 0;
	img = img_;
	dim.set(img->size[0] / 16.0, img->size[1] / 16.0); // size of letter
}

void Font::draw(char c, vect2f pos)
{
	int x = c % 16;
	int y = 15 - c / 16;

	vect2f src; src(x * dim[0], y * dim[1]);

	// draw character
	img->bind();

	// find texture coordinates
	float texX1 = src[0] / img->size[0];
	float texY1 = src[1] / img->size[1];
	float texX2 = (src[0] + dim[0]) / img->size[0];
	float texY2 = (src[1] + dim[1]) / img->size[1];

	// draw it
	glBegin(GL_QUADS);
	glTexCoord2f(texX1, texY1);
	glVertex2f(pos[0], pos[1]);

	glTexCoord2f(texX1, texY2);
	glVertex2f(pos[0], pos[1] + img->size[1] / 16.0);

	glTexCoord2f(texX2, texY2);
	glVertex2f(pos[0] + img->size[0] / 16.0, pos[1] + img->size[1] / 16.0);

	glTexCoord2f(texX2, texY1);
	glVertex2f(pos[0] + img->size[0] / 16.0, pos[1]);
	glEnd();
}

void Font::draw(string str, vect2f pos)
{
	int strLen = str.size();
	int i;
	
	int row = 0;
	int col = 0;

	// draw text
	for (i = 0; i < strLen; i++)
	{
		if (str[i] == '\t')
		{
			col += 4;
			col -= col % 4;
			continue;
		}
		else if (str[i] == '\n')
		{
			col = 0;
			row--;
			continue;
		}
		else
		{
			vect2f off; off(col * (dim[0] + spacing[0]) + (spacing[0]/2), row * (dim[1] + spacing[1]));
			draw(str[i], pos + off);
			col++;
		}
	}
}

vect2f Font::get_size(string &str)
{
	int strLen = str.size();
	int i;
	
	int row = 0;
	int col = 0;
	int maxcol = 0;

	// draw text
	for (i = 0; i < strLen; i++)
	{
		if (str[i] == '\t')
		{
			col += 4;
			col -= col % 4;
			maxcol = max(col, maxcol);
			continue;
		}
		else if (str[i] == '\n')
		{
			col = 0;
			row--;
			continue;
		}
		else
		{
			col++;
			maxcol = max(col, maxcol);
		}
	}

	vect2f ret; ret((dim[0]+spacing[0])*maxcol, (dim[1] + spacing[1])*(row+1));
	return ret;
}

vect2f Font::get_size()
{
	vect2f ret; ret(dim[0]+spacing[0], dim[1]+spacing[1]);
	return ret;
}
#else
Font::Font() {}
Font::Font(Image *img_) {}
Font::~Font() {}
void Font::load(Image *img_) {}
void Font::draw(char c, vect2f pos) {}
void Font::draw(string str, vect2f pos) {}
vect2f Font::get_size(string &str) {vect2f p; p = 0; return p;}
vect2f Font::get_size() {vect2f p; p = 0; return p;}
#endif
