#include "mesh.h"
#include "parse.h"
#include <fstream>
#include <stdlib.h>

void Mesh::load(string fn)
{
	pos.s = 0;
	uv.s = 0;
	tris.s = 0;

	ifstream in(fn.c_str());
	if (in.fail())
	{
		printf("Error: failed to load obj file\n");
		exit(1);
	}

	while (true)
	{
		string line = readLine(in);

		if (in.eof())
			break;

		Array<string> toks;
		split(toks, line, " \t\r");
		removeEmptyStrings(toks);

		if (toks.empty())
			continue;

		if (toks[0] == "v")
		{
			vect3f p;
			p[0] = atof(toks[1].c_str());
			p[1] = atof(toks[2].c_str());
			p[2] = atof(toks[3].c_str());
			pos.push_back(p);
		}
		else if (toks[0] == "vt")
		{
			vect2f p;
			p[0] = atof(toks[1].c_str());
			p[1] = atof(toks[2].c_str());
			uv.push_back(p);
		}
		else if (toks[0] == "f")
		{
			Array<string> a;
			Triangle t;

			split(a, toks[1], "/");
			t.p[0] = atoi(a[0].c_str()) - 1;
			t.t[0] = atoi(a[1].c_str()) - 1;

			for (int i = 2; i < toks.s-1; i++)
			{
				split(a, toks[i], "/");
				t.p[1] = atoi(a[0].c_str()) - 1;
				t.t[1] = atoi(a[1].c_str()) - 1;
				
				split(a, toks[i+1], "/");
				t.p[2] = atoi(a[0].c_str()) - 1;
				t.t[2] = atoi(a[1].c_str()) - 1;

				assert(t.p[0] >= 0 && t.p[0] < pos.s);
				assert(t.p[1] >= 0 && t.p[1] < pos.s);
				assert(t.p[2] >= 0 && t.p[2] < pos.s);

				assert(t.t[0] >= 0 && t.t[0] < uv.s);
				assert(t.t[1] >= 0 && t.t[1] < uv.s);
				assert(t.t[2] >= 0 && t.t[2] < uv.s);

				tris.push_back(t);
			}
		}
	}

	in.close();

	//scaleToBox();
}

void Mesh::scaleToBox()
{
	// find max,min
	vect3f t, b; // top, bottom
	t = pos[0];
	b = pos[0];
	for (int i = 0; i < pos.s; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			if (pos[i][j] < b[j])
				b[j] = pos[i][j];
			if (pos[i][j] > t[j])
				t[j] = pos[i][j];
		}
	}

	// scale
	float e = 0;
	e = max(e, t[0] - b[0]);
	e = max(e, t[1] - b[1]);
	e = max(e, t[2] - b[2]);

	for (int i = 0; i < pos.s; i++)
	{
		pos[i] -= (b + t) * .5;
		pos[i] /= e;
	}
}

float Mesh::get_area(int i)
{
	vect3f &p0 = pos[tris[i].p[0]];
	vect3f &p1 = pos[tris[i].p[1]];
	vect3f &p2 = pos[tris[i].p[2]];

	vect3f P0 = p1 - p0;
	vect3f P1 = p2 - p0;

	double pa = .5 * (P0%P1).length();

	return pa;
}
