#pragma once


// how should the edge of an image be treated?
#define MIP_CLAMP
//#define MIP_WRAP

// what method for solving matrices?
#define SOLVE_DIRECT // Directly solve using Cholesky factorization and back substitution via TAUCS
//#define SOLVE_ITERATIVE // Solve using the iterative method of Conjugate Gradients using TAUCS
//#define SOLVE_CONSTRAINED // Constrain values to be in [0,1]. Solve this quadratic programming problem using Gurobi

//#define DRAW_OPENGL
//#define CALC_ERROR




#include "vect.h"
#include "image.h"
#include "array.h"
#include "array2d.h"
#include "mesh.h"
#include "taucs_matrix.h"

struct MipStack
{
	Array<Image> tex;
	Array< Array2D<vect4f> > img;
	Image stack;
};

struct TexPoly
{
	Array<vect2f> uv;
	Array<vect3f> pos;
};

struct Texel
{
	Array<TexPoly> polys;
};

#ifdef SOLVE_CONSTRAINED
void solve_constrained(SparseTaucsMatrix &A, DenseTaucsMatrix &X, DenseTaucsMatrix &B);
#endif

void build_texels(Array<Array2D<Texel> > &texels, Mesh &mesh, int max_mip);

double get_area_obj(vect3f &p0, vect3f &p1, vect3f &p2);
double get_area_obj(TexPoly &p);

void fix_unconstrained(SparseTaucsMatrix &A, DenseTaucsMatrix &B);
float clamp01(float v);

void calc_mip_rainbow(MipStack &mip, Array2D<vect4f> &src);
void calc_mip_box_dumb(MipStack &mip, Array2D<vect4f> &src);
void calc_mip_tent_dumb(MipStack &mip, Array2D<vect4f> &src);
void calc_mip_lanczos3_dumb(MipStack &mip, Array2D<vect4f> &src);

void calc_mip_box_ignore(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh);

void calc_mip_box_smart(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh);
void calc_mip_box_smart_texels(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh, Array<Array2D<Texel> > &texels, int level_to_use);

void calc_mip_tent_smart(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh);
void calc_mip_trilinear_clamp(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh);

void build_texels(Array2D<Texel> &texels, Mesh &mesh, int max_mip);

void gutter(Array2D<vect4f> &dst, Array2D<vect4f> &src, Array2D<Texel> &texels);

void upload_texture(MipStack &mip);

template<class T>
void wrap(T &i, int v)
{
	i = (i + v) % v;
}

template<class T>
T wrap_ret(T i, int v)
{
	return (i + v) % v;
}
