#include "mip.h"
#include "global.h"
#include <GL/glew.h>

// This could be done more efficiently by using the tree structure rather than reading from the source image.

void mip_img_box_ignore(Array2D<vect4f> &dst, Array2D<vect4f> &src, Array2D<Texel> &texels, int size)
{
	dst.resize(src.s[0]/size, src.s[1]/size);

	for (int y = 0; y < dst.s[1]; y++)
	{
		for (int x = 0; x < dst.s[0]; x++)
		{
			
			double weight = 0;
			vect4f sum;
			sum = 0;
			for (int j = 0; j < size; j++)
			{
				for (int i = 0; i < size; i++)
				{
					Array<TexPoly> &p = texels(x*size + i, y*size + j).polys;

					if (p.s)
					{
						sum += src(x*size + i, y*size + j);
						weight++;
					}
				}
			}

			if (weight == 0)
				dst(x,y) = 0;
			else
				dst(x,y) = sum / weight;

		}
	}
}

void gutter(Array2D<vect4f> &dst, Array2D<vect4f> &src, Array2D<Texel> &texels)
{
	int size = 2;

	for (int y = 0; y < src.s[1]; y++)
	{
		for (int x = 0; x < src.s[0]; x++)
		{
			
			for (int j = 0; j < size; j++)
			{
				for (int i = 0; i < size; i++)
				{
					Array<TexPoly> &p = texels(x*size + i, y*size + j).polys;

					if (p.s == 0)
					{
						//// nearest neighbor filter
						//dst(x*size + i, y*size + j) = src(x, y);

						// bilinear filter
#ifdef MIP_CLAMP
						dst(x*size + i, y*size + j) = (1./16.) * ( src.get_clamp(x+i, y+j) + src.get_clamp(x-1+i, y+j) + src.get_clamp(x+i, y-1+j) + src.get_clamp(x-1+i, y-1+j) )
													+ (2./16.) * ( src.get_clamp(x+i, y) + src.get_clamp(x-1+i, y) )
													+ (2./16.) * ( src.get_clamp(x, y+j) + src.get_clamp(x, y-1+j) )
													+ (4./16.) * ( src.get_clamp(x, y) );
#else
						dst(x*size + i, y*size + j) = (1./16.) * ( src.get_wrap(x+i, y+j) + src.get_wrap(x-1+i, y+j) + src.get_wrap(x+i, y-1+j) + src.get_wrap(x-1+i, y-1+j) )
													+ (2./16.) * ( src.get_wrap(x+i, y) + src.get_wrap(x-1+i, y) )
													+ (2./16.) * ( src.get_wrap(x, y+j) + src.get_wrap(x, y-1+j) )
													+ (4./16.) * ( src.get_wrap(x, y) );
#endif
					}
				}
			}
		}
	}
}

void calc_mip_box_ignore_texels(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh, Array<Array2D<Texel> > &texels, int level_to_use)
{
	// build mip images
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

#if 1
	mip.img[0] = src;
	mip.tex[0].load(mip.img[0]);

	for (int i = 1; i < g.num_mip_levels; i++)
	{
		mip_img_box_ignore(mip.img[i], src, texels[level_to_use], 1 << i);
	}
#else
	for (int i = 0; i < g.num_mip_levels; i++)
	{
		mip_img_box_ignore(mip.img[i], src, texels[level_to_use], 1 << i);
	}
#endif
	
	// gutter
	for (int i = g.num_mip_levels - 1; i > 1; i--)
	{
		gutter(mip.img[i-1], mip.img[i], texels[i-1]);
	}

	// upload to opengl
	upload_texture(mip);
}


void calc_mip_box_ignore(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh)
{
	// build polygon texel lists
	Array<Array2D<Texel> > texels;
	build_texels(texels, mesh, g.num_mip_levels-1);
	
	// calc
	calc_mip_box_ignore_texels(mip, src, mesh, texels, 0);
}
