#include "mip.h"
#include "global.h"
#include <GL/glew.h>

// This could be done more efficiently by using the tree structure rather than reading from the source image.

void mip_img_box_smart(Array2D<vect4f> &dst, Array2D<vect4f> &src, Array2D<Texel> &texels, int size)
{
	dst.resize(src.s[0]/size, src.s[1]/size);

	for (int y = 0; y < dst.s[1]; y++)
	{
		for (int x = 0; x < dst.s[0]; x++)
		{
			
			double weight = 0;
			vect4f sum;
			sum = 0;
			for (int j = 0; j < size; j++)
			{
				for (int i = 0; i < size; i++)
				{
					double w = 0;
					Array<TexPoly> &p = texels(x*size + i, y*size + j).polys;
					for (int k = 0; k < p.s; k++)
					{
						TexPoly &pk = p[k];
						//w += get_area_uv(pk) * pk.stretch;
						w += get_area_obj(pk);

						//double ao = get_area_obj(pk) * 256 * 256;
						//double at = get_area_uv(pk);
						//double ratio = ao / at;
						//printf("obj = %f, tex = %f, ratio = %f\n", ao, at, ratio);
					}

					sum += src(x*size + i, y*size + j) * w;
					weight += w;
				}
			}

			if (weight == 0)
				dst(x,y) = 0;
			else
				dst(x,y) = sum / weight;

		}
	}
}

void calc_mip_box_smart_texels(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh, Array<Array2D<Texel> > &texels, int level_to_use)
{
	// build mip images
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

#if 1
	mip.img[0] = src;
	mip.tex[0].load(mip.img[0]);

	for (int i = 1; i < g.num_mip_levels; i++)
	{
		mip_img_box_smart(mip.img[i], src, texels[level_to_use], 1 << i);
	}
#else
	for (int i = 0; i < g.num_mip_levels; i++)
	{
		mip_img_box_smart(mip.img[i], src, texels[level_to_use], 1 << i);
	}
#endif
	
	// gutter
	for (int i = g.num_mip_levels - 1; i > 1; i--)
	{
		gutter(mip.img[i-1], mip.img[i], texels[i-1]);
	}

	// upload to opengl
	upload_texture(mip);
}


void calc_mip_box_smart(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh)
{
	// build polygon texel lists
	Array<Array2D<Texel> > texels;
	build_texels(texels, mesh, g.num_mip_levels-1);
	
	// calc
	calc_mip_box_smart_texels(mip, src, mesh, texels, 0);
}
