#include "mip.h"
#include "global.h"
#include <GL/glew.h>

static double l3(double x)
{
	if (x <= -3 || x >= 3)
		return 0;
	else if (fabs(x) < .0001)
		return 1;
	else
		return sin(pi*x) / (pi*x);
		//return 3*sin(pi*x/3) * sin(pi*x) / (pi*pi*x*x);
}

static double filter(double x, double y)
{
	return l3(x)*l3(y);
}

static void mip_img_lanczos3_dumb(Array2D<vect4f> &dst, Array2D<vect4f> &src, int size)
{
	dst.resize(src.s[0]/size, src.s[1]/size);

	for (int y = 0; y < dst.s[1]; y++)
	{
		for (int x = 0; x < dst.s[0]; x++)
		{
			
			double sum_w = 0;
			vect4f sum;
			sum = 0;
			for (int j = -5*size/2; j < 7*size/2; j++)
			{
				for (int i = -5*size/2; i < 7*size/2; i++)
				{
					int sx = x*size + i;
					int sy = y*size + j;
#ifdef MIP_WRAP
					wrap(sx, src.s[0]);
					wrap(sy, src.s[1]);
#endif
					if (sx >= 0 && sx < src.s[0] && sy >= 0 && sy < src.s[1])
					{
						double w = filter((i - .5*(size - 1))/(double)size, (j - .5*(size - 1))/(double)size);
						sum += src(sx, sy) * w;
						sum_w += w;
					}
				}
			}

			dst(x,y)[0] = clamp01(sum[0] / sum_w);
			dst(x,y)[1] = clamp01(sum[1] / sum_w);
			dst(x,y)[2] = clamp01(sum[2] / sum_w);
		}
	}
}

void calc_mip_lanczos3_dumb(MipStack &mip, Array2D<vect4f> &src)
{
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

	mip.img[0] = src;
	mip.tex[0].load(src);

	for (int i = 1; i < g.num_mip_levels; i++)
	{
		mip_img_lanczos3_dumb(mip.img[i], src, 1 << i);
	}

	// upload to opengl
	upload_texture(mip);
}