#include "mip.h"
#include "global.h"
#include <GL/glew.h>

static double tent(double x)
{
	if (x <= -1 || x >= 1)
		return 0;
	else if (x < 0)
		return 1 + x;
	else
		return 1 - x;
}

static double filter(double x, double y)
{
	return tent(x)*tent(y);
}

static void mip_img_tent_dumb(Array2D<vect4f> &dst, Array2D<vect4f> &src, int size)
{
	dst.resize(src.s[0]/size, src.s[1]/size);

	for (int y = 0; y < dst.s[1]; y++)
	{
		for (int x = 0; x < dst.s[0]; x++)
		{
			
			double sum_w = 0;
			vect4f sum;
			sum = 0;
			for (int j = -size/2; j < 3*size/2; j++)
			{
				for (int i = -size/2; i < 3*size/2; i++)
				{
					int sx = x*size + i;
					int sy = y*size + j;
#ifdef MIP_WRAP
					wrap(sx, src.s[0]);
					wrap(sy, src.s[1]);
#endif
					if (sx >= 0 && sx < src.s[0] && sy >= 0 && sy < src.s[1])
					{
						double w = filter((i - .5*(size - 1))/(double)size, (j - .5*(size - 1))/(double)size);
						sum += src(sx, sy) * w;
						sum_w += w;
					}
				}
			}
			dst(x,y) = sum / sum_w;

		}
	}
}

void calc_mip_tent_dumb(MipStack &mip, Array2D<vect4f> &src)
{
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

	mip.img[0] = src;
	mip.tex[0].load(src);

	for (int i = 1; i < g.num_mip_levels; i++)
	{
		mip_img_tent_dumb(mip.img[i], src, 1 << i);
	}

	// upload to opengl
	upload_texture(mip);
}