#include "mip.h"
#include "global.h"
#include "timer.h"
#include "taucs_matrix.h"
#include <GL/glew.h>

#include <omp.h>

static __forceinline double Power(double v, int p)
{
	if (p == 0)
		return 1;
	else if (p == 1)
		return v;
	else if (p == 2)
		return v*v;
	else if (p == 3)
		return v*v*v;
	else if (p == 4)
		return v*v*v*v;
	else if (p == 5)
		return v*v*v*v*v;
	else if (p == 6)
		return v*v*v*v*v*v;
	else
	{
		assert(false);
		return 0;
	}
}

static double eval_bilinear(double p00, double p01, 
							double p10, double p11, 
							double p20, double p21, 
							double A0, double A1, double B0, double B1, 
							double a0, double a1, double b0, double b1)
{
	double aa0 = a0 * A0;
	double aa1 = A1 * a0 + A0 * a1;
	double aa2 = a1 * A1;
	double bb0 = b0 * B0;
	double bb1 = B1 * b0 + B0 * b1;
	double bb2 = b1 * B1;

	return (3*a0*(10*A0*(2*b0*(3*B0 + B1*(p01 + p11 + p21)) + b1*(2*B0*(p01 + p11 + p21) + B1*(Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + p01*(p11 + p21)))) + 
        A1*(5*b0*(4*B0*(p00 + p10 + p20) + B1*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21))) + 
           b1*(5*B0*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21)) + 
              2*B1*(3*p10*Power(p11,2) + Power(p11,2)*p20 + Power(p01,2)*(p10 + p20) + 2*p10*p11*p21 + 2*p11*p20*p21 + p10*Power(p21,2) + 3*p20*Power(p21,2) + p01*(2*p10*p11 + p11*p20 + p10*p21 + 2*p20*p21) + 
                 p00*(3*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 2*p01*(p11 + p21)))))) + 
     a1*(3*A0*(5*b0*(4*B0*(p00 + p10 + p20) + B1*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21))) + 
           b1*(5*B0*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21)) + 
              2*B1*(3*p10*Power(p11,2) + Power(p11,2)*p20 + Power(p01,2)*(p10 + p20) + 2*p10*p11*p21 + 2*p11*p20*p21 + p10*Power(p21,2) + 3*p20*Power(p21,2) + p01*(2*p10*p11 + p11*p20 + p10*p21 + 2*p20*p21) + 
                 p00*(3*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 2*p01*(p11 + p21))))) + 
        2*A1*(3*b0*(5*B0*(Power(p00,2) + Power(p10,2) + p10*p20 + Power(p20,2) + p00*(p10 + p20)) + 
              B1*(3*Power(p10,2)*p11 + 2*p10*p11*p20 + p11*Power(p20,2) + p01*(Power(p10,2) + p10*p20 + Power(p20,2)) + Power(p10,2)*p21 + 2*p10*p20*p21 + 3*Power(p20,2)*p21 + Power(p00,2)*(3*p01 + p11 + p21) + 
                 p00*(2*p10*p11 + p11*p20 + 2*p01*(p10 + p20) + p10*p21 + 2*p20*p21))) + 
           b1*(3*B0*(3*Power(p10,2)*p11 + 2*p10*p11*p20 + p11*Power(p20,2) + p01*(Power(p10,2) + p10*p20 + Power(p20,2)) + Power(p10,2)*p21 + 2*p10*p20*p21 + 3*Power(p20,2)*p21 + 
                 Power(p00,2)*(3*p01 + p11 + p21) + p00*(2*p10*p11 + p11*p20 + 2*p01*(p10 + p20) + p10*p21 + 2*p20*p21)) + 
              B1*(6*Power(p10,2)*Power(p11,2) + 3*p10*Power(p11,2)*p20 + Power(p11,2)*Power(p20,2) + Power(p01,2)*(Power(p10,2) + p10*p20 + Power(p20,2)) + 3*Power(p10,2)*p11*p21 + 4*p10*p11*p20*p21 + 
                 3*p11*Power(p20,2)*p21 + Power(p10,2)*Power(p21,2) + 3*p10*p20*Power(p21,2) + 6*Power(p20,2)*Power(p21,2) + 
                 Power(p00,2)*(6*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 3*p01*(p11 + p21)) + p01*(2*p10*p20*(p11 + p21) + Power(p10,2)*(3*p11 + p21) + Power(p20,2)*(p11 + 3*p21)) + 
                 p00*(3*Power(p01,2)*(p10 + p20) + p10*(3*Power(p11,2) + 2*p11*p21 + Power(p21,2)) + p20*(Power(p11,2) + 2*p11*p21 + 3*Power(p21,2)) + 2*p01*(p10*(2*p11 + p21) + p20*(p11 + 2*p21))))))))/360.;
}

void mip_img_tent_smart(Array2D<vect4f> &dst, Array2D<vect4f> &src, Array< Array2D<Texel> > &texels, int size, int mip_level)
{
	double t0 = get_time();

	dst.resize(src.s[0]/size, src.s[1]/size);

	int s2 = 2*size;
	double s2d = s2;

	//** Build A matrix
	SparseTaucsMatrix A;
	A.resize(dst.s[0]*dst.s[1]);
	A.set_zero();

#pragma omp parallel
	{
		int nthreads = omp_get_num_threads();
		int tid = omp_get_thread_num();

		SparseTaucsMatrix Alocal;
		if (nthreads > 1)
		{
			Alocal.resize(dst.s[0]*dst.s[1]);
			Alocal.set_zero();
		}

		// for all texels
		for (int y = tid; y < texels[mip_level].s[1]; y += nthreads)
		{
			int Oy = (y+1)%2;

			for (int x = 0; x < texels[mip_level].s[0]; x++)
			{
				int Ox = (x+1)%2;

				Array<TexPoly> &p = texels[mip_level](x, y).polys;

				// for all rows of matrix
				for (int rj = 0; rj <= 1; rj++)
				{
					double B1 = (2*rj - 1) / 2.0;
					double B0 = 1 - rj + B1*Oy;

					int Ry = ((y+1)/2 + rj - 1);

#ifdef MIP_CLAMP
					if (Ry < 0)
						Ry = 0;
					else if (Ry >= dst.s[1])
						Ry = dst.s[1] - 1;
#else
					wrap(Ry, dst.s[1]);
#endif

					for (int ri = 0; ri <= 1; ri++)
					{
						int Rx = ((x+1)/2 + ri - 1);

#ifdef MIP_CLAMP
						if (Rx < 0)
							Rx = 0;
						else if (Rx >= dst.s[0])
							Rx = dst.s[0] - 1;
#else
						wrap(Rx, dst.s[0]);
#endif

						int R = Ry*dst.s[0] + Rx;

						double A1 = (2*ri - 1) / 2.0;
						double A0 = 1 - ri + A1*Ox;

						// for all columns of matrix
						for (int cj = 0; cj <= 1; cj++)
						{
							double b1 = (2*cj - 1) / 2.0;
							double b0 = 1 - cj + b1*Oy;

							int Cy = ((y+1)/2 + cj - 1);

#ifdef MIP_CLAMP
							if (Cy < 0)
								Cy = 0;
							else if (Cy >= dst.s[1])
								Cy = dst.s[1] - 1;
#else
							wrap(Cy, dst.s[1]);
#endif
							for (int ci = 0; ci <= 1; ci++)
							{
								int Cx = ((x+1)/2 + ci - 1);

#ifdef MIP_CLAMP
								if (Cx < 0)
									Cx = 0;
								else if (Cx >= dst.s[0])
									Cx = dst.s[0] - 1;
#else
								wrap(Cx, dst.s[0]);
#endif

								int C = Cy*dst.s[0] + Cx;

								double a1 = (2*ci - 1) / 2.0;
								double a0 = 1 - ci + a1*Ox;

								// for all triangles
								double Av = 0;
								for (int k = 0; k < p.s; k++)
								{
									double p00 = p[k].uv[0][0];
									double p01 = p[k].uv[0][1];

									for (int l = 2; l < p[k].uv.s; l++)
									{
										double p10 = p[k].uv[l-1][0];
										double p11 = p[k].uv[l-1][1];

										double p20 = p[k].uv[l][0];
										double p21 = p[k].uv[l][1];

										double coeff = get_area_obj(p[k].pos[0], p[k].pos[l-1], p[k].pos[l]) * texels[mip_level].s[0]*texels[mip_level].s[1];
										Av += coeff *
											eval_bilinear(
											p00, p01, p10, p11, p20, p21,
											A0, A1, B0, B1,
											a0, a1, b0, b1
											);
									}
								}

								// add entry to A and B
								if (Av != 0)
								{
									if (nthreads > 1)
										Alocal(R,C) += Av;
									else
										A(R,C) += Av;
								}
							}
						}

					}
				}

			}
		}

		// reduce Alocal into A
#pragma omp critical
		if (nthreads > 1)
		{
			A += Alocal;
		}
	}

	//** Build B matrix
	DenseTaucsMatrix B;
	B.resize(dst.s[0]*dst.s[1], 3);
	B.set_zero();

#pragma omp parallel
	{
		int nthreads = omp_get_num_threads();
		int tid = omp_get_thread_num();

		DenseTaucsMatrix Blocal;
		if (nthreads > 1)
		{
			Blocal.resize(dst.s[0]*dst.s[1], 3);
			Blocal.set_zero();
		}

		// for all texels
		for (int y = tid; y < texels[0].s[1]; y += nthreads)
		{
			int Oy = (y+size)%s2;
			int Osy = (y+1)%2;

			for (int x = 0; x < texels[0].s[0]; x++)
			{
				int Ox = (x+size)%s2;
				int Osx = (x+1)%2;
				Array<TexPoly> &p = texels[0](x, y).polys;

				// for all rows of matrix
				for (int rj = 0; rj <= 1; rj++)
				{
					double B1 = (2*rj - 1) / s2d;
					double B0 = 1 - rj + B1*Oy;

					int Ry = ((y+size)/s2 + rj - 1);

#ifdef MIP_CLAMP
					if (Ry < 0)
						Ry = 0;
					else if (Ry >= dst.s[1])
						Ry = dst.s[1] - 1;
#else
					wrap(Ry, dst.s[1]);
#endif
					for (int ri = 0; ri <= 1; ri++)
					{
						int Rx = ((x+size)/s2 + ri - 1);

#ifdef MIP_CLAMP
						if (Rx < 0)
							Rx = 0;
						else if (Rx >= dst.s[0])
							Rx = dst.s[0] - 1;
#else
						wrap(Rx, dst.s[0]);
#endif

						int R = Ry*dst.s[0] + Rx;

						double A1 = (2*ri - 1) / s2d;
						double A0 = 1 - ri + A1*Ox;

						// for all columns of matrix
						for (int cj = 0; cj <= 1; cj++)
						{
							double bs1 = (2*cj - 1) / 2.0;
							double bs0 = 1 - cj + bs1*Osy;
							int CBy = ((y+1)/2 + cj - 1);

#ifdef MIP_CLAMP
							if (CBy < 0)
								CBy = 0;
							else if (CBy >= src.s[1])
								CBy = src.s[1] - 1;
#else
							wrap(CBy, src.s[1]);
#endif
							for (int ci = 0; ci <= 1; ci++)
							{
								int CBx = ((x+1)/2 + ci - 1);

#ifdef MIP_CLAMP
								if (CBx < 0)
									CBx = 0;
								else if (CBx >= src.s[0])
									CBx = src.s[0] - 1;
#else
								wrap(CBx, src.s[0]);
#endif

								double as1 = (2*ci - 1) / 2.0;
								double as0 = 1 - ci + as1*Osx;

								// for all triangles
								double Bv = 0;
								for (int k = 0; k < p.s; k++)
								{
									double p00 = p[k].uv[0][0];
									double p01 = p[k].uv[0][1];

									for (int l = 2; l < p[k].uv.s; l++)
									{
										double p10 = p[k].uv[l-1][0];
										double p11 = p[k].uv[l-1][1];

										double p20 = p[k].uv[l][0];
										double p21 = p[k].uv[l][1];

										double coeff = get_area_obj(p[k].pos[0], p[k].pos[l-1], p[k].pos[l]) * texels[mip_level].s[0]*texels[mip_level].s[1];
										Bv += coeff *
											eval_bilinear(
											p00, p01, p10, p11, p20, p21,
											A0, A1, B0, B1,
											as0, as1, bs0, bs1
											);
									}
								}

								// add entry to A and B
								vect4f &sc = src(CBx, CBy);
								if (nthreads > 1)
								{
									Blocal(R,0) += Bv * sc[0];
									Blocal(R,1) += Bv * sc[1];
									Blocal(R,2) += Bv * sc[2];
								}
								else
								{
									B(R,0) += Bv * sc[0];
									B(R,1) += Bv * sc[1];
									B(R,2) += Bv * sc[2];
								}
							}
						}

					}
				}

			}
		}

		// reduce Blocal into B
#pragma omp critical
		if (nthreads > 1)
		{
			B += Blocal;
		}

	}

	// a model may not use all of the texture, so constrain unused texels
	fix_unconstrained(A, B);

	// solve
	DenseTaucsMatrix X;
	A.make_ccs();
	A.data.clear();

#ifdef SOLVE_DIRECT
	A.factorize();
	A.solve(X, B);
#elif defined(SOLVE_ITERATIVE)
	A.solve_cg(X, B);
#elif defined(SOLVE_CONSTRAINED)
	solve_constrained(A, X, B);
#endif

	// store solution
	for (int j = 0; j < dst.s[1]; j++)
	{
		for (int i = 0; i < dst.s[0]; i++)
		{
			dst(i,j)[0] = clamp01(X(j*dst.s[0] + i, 0));
			dst(i,j)[1] = clamp01(X(j*dst.s[0] + i, 1));
			dst(i,j)[2] = clamp01(X(j*dst.s[0] + i, 2));
			dst(i,j)[3] = 1;
		}
	}
	double t5 = get_time();

	printf("--== Tent Smart downscale by factor %d ==--\n", size);
	printf("time = %f\n", t5 - t0);
}

void calc_mip_tent_smart(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh)
{
	// build polygon texel lists
	Array<Array2D<Texel> > texels;
	build_texels(texels, mesh, g.num_mip_levels);

	// build mip images
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

	mip.img[0] = src;
	mip.tex[0].load(mip.img[0]);

	for (int i = 1; i < g.num_mip_levels; i++)
	{
		mip_img_tent_smart(mip.img[i], src, texels, 1 << i, i);
	}

	// upload to opengl
	upload_texture(mip);
}
