#include "mip.h"
#include "global.h"
#include "timer.h"
#include "taucs_matrix.h"
#include <GL/glew.h>

#include <omp.h>

static __forceinline double Power(double v, int p)
{
	if (p == 0)
		return 1;
	else if (p == 1)
		return v;
	else if (p == 2)
		return v*v;
	else if (p == 3)
		return v*v*v;
	else if (p == 4)
		return v*v*v*v;
	else if (p == 5)
		return v*v*v*v*v;
	else if (p == 6)
		return v*v*v*v*v*v;
	else
	{
		assert(false);
		return 0;
	}
}

static double eval_trilinear(double p00, double p01, 
							 double p10, double p11, 
							 double p20, double p21, 
							 double A0, double A1, double B0, double B1, double C0, double C1, 
							 double a0, double a1, double b0, double b1, double c0, double c1)
{
	double aa0 = a0 * A0;
	double aa1 = A1 * a0 + A0 * a1;
	double aa2 = a1 * A1;
	double bb0 = b0 * B0;
	double bb1 = B1 * b0 + B0 * b1;
	double bb2 = b1 * B1;

	return -((6*c0*C0 + 3*C0*c1 + 3*c0*C1 + 2*c1*C1)*(-3*a0*(10*A0*(2*b0*(3*B0 + B1*(p01 + p11 + p21)) + b1*(2*B0*(p01 + p11 + p21) + B1*(Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + p01*(p11 + p21)))) + 
           A1*(5*b0*(4*B0*(p00 + p10 + p20) + B1*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21))) + 
              b1*(5*B0*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21)) + 
                 2*B1*(3*p10*Power(p11,2) + Power(p11,2)*p20 + Power(p01,2)*(p10 + p20) + 2*p10*p11*p21 + 2*p11*p20*p21 + p10*Power(p21,2) + 3*p20*Power(p21,2) + 
                    p01*(2*p10*p11 + p11*p20 + p10*p21 + 2*p20*p21) + p00*(3*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 2*p01*(p11 + p21)))))) - 
        a1*(3*A0*(5*b0*(4*B0*(p00 + p10 + p20) + B1*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21))) + 
              b1*(5*B0*(2*p10*p11 + p11*p20 + p01*(p10 + p20) + p10*p21 + 2*p20*p21 + p00*(2*p01 + p11 + p21)) + 
                 2*B1*(3*p10*Power(p11,2) + Power(p11,2)*p20 + Power(p01,2)*(p10 + p20) + 2*p10*p11*p21 + 2*p11*p20*p21 + p10*Power(p21,2) + 3*p20*Power(p21,2) + 
                    p01*(2*p10*p11 + p11*p20 + p10*p21 + 2*p20*p21) + p00*(3*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 2*p01*(p11 + p21))))) + 
           2*A1*(3*b0*(5*B0*(Power(p00,2) + Power(p10,2) + p10*p20 + Power(p20,2) + p00*(p10 + p20)) + 
                 B1*(3*Power(p10,2)*p11 + 2*p10*p11*p20 + p11*Power(p20,2) + p01*(Power(p10,2) + p10*p20 + Power(p20,2)) + Power(p10,2)*p21 + 2*p10*p20*p21 + 3*Power(p20,2)*p21 + 
                    Power(p00,2)*(3*p01 + p11 + p21) + p00*(2*p10*p11 + p11*p20 + 2*p01*(p10 + p20) + p10*p21 + 2*p20*p21))) + 
              b1*(3*B0*(3*Power(p10,2)*p11 + 2*p10*p11*p20 + p11*Power(p20,2) + p01*(Power(p10,2) + p10*p20 + Power(p20,2)) + Power(p10,2)*p21 + 2*p10*p20*p21 + 3*Power(p20,2)*p21 + 
                    Power(p00,2)*(3*p01 + p11 + p21) + p00*(2*p10*p11 + p11*p20 + 2*p01*(p10 + p20) + p10*p21 + 2*p20*p21)) + 
                 B1*(6*Power(p10,2)*Power(p11,2) + 3*p10*Power(p11,2)*p20 + Power(p11,2)*Power(p20,2) + Power(p01,2)*(Power(p10,2) + p10*p20 + Power(p20,2)) + 3*Power(p10,2)*p11*p21 + 4*p10*p11*p20*p21 + 
                    3*p11*Power(p20,2)*p21 + Power(p10,2)*Power(p21,2) + 3*p10*p20*Power(p21,2) + 6*Power(p20,2)*Power(p21,2) + 
                    Power(p00,2)*(6*Power(p01,2) + Power(p11,2) + p11*p21 + Power(p21,2) + 3*p01*(p11 + p21)) + p01*(2*p10*p20*(p11 + p21) + Power(p10,2)*(3*p11 + p21) + Power(p20,2)*(p11 + 3*p21)) + 
                    p00*(3*Power(p01,2)*(p10 + p20) + p10*(3*Power(p11,2) + 2*p11*p21 + Power(p21,2)) + p20*(Power(p11,2) + 2*p11*p21 + 3*Power(p21,2)) + 2*p01*(p10*(2*p11 + p21) + p20*(p11 + 2*p21)))))))))/
   2160.;
}

// CLAMP TOP AND BOTTOM
struct TrilinearClampCalcData
{
	int i, x, y;
	int ri, rj;
	int ci, cj;

	int R, Rx, Ry;
	int CBx, CBy;

	int size;

	int Ox, Oy;
	int OSx, OSy;
	int Osx, Osy;

	double A1, A0, B1, B0;
	double C1, C0;

	double as0, as1, bs0, bs1;
	double a1, a0, b1, b0;
	double c1, c0;

	double p00, p01;
	double p10, p11;
	double p20, p21;

	vect3f pos0, pos1, pos2;

	vect4f avg_color;
};

__forceinline static void write_entry_B(DenseTaucsMatrix &B, Array< Array2D<Texel> > &texels,
										Array2D<vect4f> &src, Array<int> &offsets, 
										TrilinearClampCalcData &d, int level)
{
	int S = 1 << (d.i+level);

	d.OSx = (d.x+S)%(2*S);
	d.OSy = (d.y+S)%(2*S);

	d.a1 = (2*d.ci - 1) / (double)(2*S);
	d.a0 = 1 - d.ci + d.a1*d.OSx;
	d.b1 = (2*d.cj - 1) / (double)(2*S);
	d.b0 = 1 - d.cj + d.b1*d.OSy;

	double Bv = get_area_obj(d.pos0, d.pos1, d.pos2) * texels[0].data_size * eval_trilinear(
		d.p00, d.p01, d.p10, d.p11, d.p20, d.p21,
		d.A0, d.A1, d.B0, d.B1, d.C0, d.C1,
		d.as0, d.as1, d.bs0, d.bs1, d.c0, d.c1
		);

	//** write values
	vect4f &sc = src(d.CBx, d.CBy);
	B(d.R, 0) += Bv * sc[0];
	B(d.R, 1) += Bv * sc[1];
	B(d.R, 2) += Bv * sc[2];
}

// fast version
__forceinline static void write_entry_A(SparseTaucsMatrix &A, DenseTaucsMatrix &B, Array< Array2D<Texel> > &texels,
										Array2D<vect4f> &src, Array<int> &offsets, 
										TrilinearClampCalcData &d, int level)
{
	int S = 1 << (1+level);

	d.OSx = (d.x+S)%(2*S);
	d.OSy = (d.y+S)%(2*S);

	d.a1 = (2*d.ci - 1) / (double)(2*S);
	d.a0 = 1 - d.ci + d.a1*d.OSx;
	d.b1 = (2*d.cj - 1) / (double)(2*S);
	d.b0 = 1 - d.cj + d.b1*d.OSy;

	double Av = get_area_obj(d.pos0, d.pos1, d.pos2) * texels[0].data_size * eval_trilinear(
		d.p00, d.p01, d.p10, d.p11, d.p20, d.p21,
		d.A0, d.A1, d.B0, d.B1, d.C0, d.C1,
		d.a0, d.a1, d.b0, d.b1, d.c0, d.c1
		);

	//** write values
	if (d.i+level <= 0)
	{
		// add clamped contribution to B
		vect4f &sc = src(d.CBx, d.CBy);
		B(d.R, 0) -= Av * sc[0];
		B(d.R, 1) -= Av * sc[1];
		B(d.R, 2) -= Av * sc[2];
	}
	else if (d.i+level >= g.num_mip_levels-1)
	{
		// add clamped contribution to B
		//		vect4f &sc = src(d.CBx, d.CBy);
		B(d.R, 0) -= Av * d.avg_color[0];
		B(d.R, 1) -= Av * d.avg_color[1];
		B(d.R, 2) -= Av * d.avg_color[2];
	}
	else
	{
		//** row and col in matrix to stick it in

		// column in A
		int Cx = ((d.x+S)/(2*S) + d.ci - 1);
		int Cy = ((d.y+S)/(2*S) + d.cj - 1);

#ifdef MIP_CLAMP
		if (Cx < 0)
			Cx = 0;
		else if (Cx >= (texels[d.i].s[0]/S))
			Cx = (texels[d.i].s[0]/S) - 1;

		if (Cy < 0)
			Cy = 0;
		else if (Cy >= (texels[d.i].s[1]/S))
			Cy = (texels[d.i].s[1]/S) - 1;
#else

		wrap(Cx, texels[d.i].s[0]/S);
		wrap(Cy, texels[d.i].s[1]/S);
#endif
		int C = Cy*(texels[d.i].s[0]/S) + Cx + offsets[d.i+level];

		A(C, d.R) += Av;
	}
}


void calc_mip_trilinear_clamp(MipStack &mip, Array2D<vect4f> &src, Mesh &mesh)
{
	// build polygon texel lists
	Array<Array2D<Texel> > texels;
	build_texels(texels, mesh, g.num_mip_levels);

	// build mip images
	mip.img.resize(g.num_mip_levels);
	mip.tex.resize(g.num_mip_levels);

	Array<int> offsets;
	int totalsize = 0;
	{
		offsets.reserve(g.num_mip_levels-1);

		offsets.push_back(0); // mip level 0, which is clamped and thus, not solved for

		for (int i = 1; i < g.num_mip_levels-1; i++)
		{
			int size = 1 << i;
			offsets.push_back(totalsize);
			totalsize += src.s[0]*src.s[1] / (size*size);
		}

		offsets.push_back(totalsize + 1000); // mip level n, which is clamped and thus, not solved for
	}

	//*** BUILD A GUESS FOR THE SOLUTION
	DenseTaucsMatrix X;
	MipStack guess;

	calc_mip_box_smart_texels(guess, src, mesh, texels, 1);
	//calc_mip_box_smart(guess, src, mesh);

	X.resize(totalsize, 3);

	for (int k = 1; k < g.num_mip_levels-1; k++)
	{
		Array2D<vect4f> &box = guess.img[k];

		for (int j = 0; j < box.s[1]; j++)
		{
			for (int i = 0; i < box.s[0]; i++)
			{
				for (int c = 0; c < 3; c++)
				{
					X(j*box.s[0] + i + offsets[k], c) = box(i,j)[c];
				}
			}
		}
	}

	//*** SOLVE THE LINEAR SYSTEM
	double t0 = get_time();

	//** Build A and B matrix
	SparseTaucsMatrix A;
	A.resize(totalsize);
	A.set_zero();

	DenseTaucsMatrix B;
	B.resize(totalsize, 3);
	B.set_zero();

#pragma omp parallel
	{
		int nthreads = omp_get_num_threads();
		int tid = omp_get_thread_num();

		SparseTaucsMatrix Alocal;
		DenseTaucsMatrix Blocal;

		if (nthreads > 1)
		{
			Alocal.resize(totalsize);
			Alocal.set_zero();
			Blocal.resize(totalsize, 3);
			Blocal.set_zero();
		}

		TrilinearClampCalcData d;
		d.avg_color = guess.img[g.num_mip_levels-1].data[0];

		// Build A fast version
		for (d.i = 1; d.i < g.num_mip_levels-1; d.i++)
		{
			d.size = 2;

			// for all texels
			for (d.y = tid; d.y < texels[d.i-1].s[1]; d.y += nthreads)
			{
				d.Oy = (d.y+d.size)%(2*d.size);

				for (d.x = 0; d.x < texels[d.i-1].s[0]; d.x++)
				{
					d.Ox = (d.x+d.size)%(2*d.size);

					Array<TexPoly> &p = texels[d.i-1](d.x, d.y).polys;

					// for all rows of matrix
					for (d.rj = 0; d.rj <= 1; d.rj++)
					{
						d.B1 = (2*d.rj - 1) / (double)(2*d.size);
						d.B0 = 1 - d.rj + d.B1*d.Oy;
						d.Ry = ((d.y+d.size)/(2*d.size) + d.rj - 1);

#ifdef MIP_CLAMP
						if (d.Ry < 0)
							d.Ry = 0;
						else if (d.Ry >= (texels[d.i].s[1]/d.size))
							d.Ry = (texels[d.i].s[1]/d.size) - 1;
#else
						wrap(d.Ry, texels[d.i].s[1]/d.size);
#endif
						for (d.ri = 0; d.ri <= 1; d.ri++)
						{

							d.A1 = (2*d.ri - 1) / (double)(2*d.size);
							d.A0 = 1 - d.ri + d.A1*d.Ox;

							d.Rx = ((d.x+d.size)/(2*d.size) + d.ri - 1);

#ifdef MIP_CLAMP
							if (d.Rx < 0)
								d.Rx = 0;
							else if (d.Rx >= (texels[d.i].s[0]/d.size))
								d.Rx = (texels[d.i].s[0]/d.size) - 1;
#else
							wrap(d.Rx, texels[d.i].s[0]/d.size);
#endif
							d.R = d.Ry*(texels[d.i].s[0]/d.size) + d.Rx + offsets[d.i];

							// for all columns of matrix
							for (d.cj = 0; d.cj <= 1; d.cj++)
							{
								d.CBy = ((d.y+1)/2 + d.cj - 1);

#ifdef MIP_CLAMP
								if (d.CBy < 0)
									d.CBy = 0;
								else if (d.CBy >= src.s[1])
									d.CBy = src.s[1] - 1;
#else
								wrap(d.CBy, src.s[1]);
#endif
								for (d.ci = 0; d.ci <= 1; d.ci++)
								{
									d.CBx = ((d.x+1)/2 + d.ci - 1);

#ifdef MIP_CLAMP
									if (d.CBx < 0)
										d.CBx = 0;
									else if (d.CBx >= src.s[0])
										d.CBx = src.s[0] - 1;
#else
									wrap(d.CBx, src.s[0]);
#endif

									// for all triangles
									for (int k = 0; k < p.s; k++)
									{
										d.p00 = p[k].uv[0][0];
										d.p01 = p[k].uv[0][1];
										d.pos0 = p[k].pos[0];

										d.p10 = p[k].uv[1][0];
										d.p11 = p[k].uv[1][1];
										d.pos1 = p[k].pos[1];
										for (int l = 2; l < p[k].uv.s; l++)
										{
											d.p20 = p[k].uv[l][0];
											d.p21 = p[k].uv[l][1];
											d.pos2 = p[k].pos[l];

											//** add entry to A and B
#if 0
											// bilinear
											if (nthreads > 1)
											{
												d.C0 = 1; d.C1 = 0; d.c0 = 1; d.c1 = 0;
												write_entry_A(Alocal, Blocal, texels, src, offsets, d, 0);
											}
											else
											{
												d.C0 = 1; d.C1 = 0; d.c0 = 1; d.c1 = 0;
												write_entry_A(A, B, texels, src, offsets, d, 0);
											}
#else
											if (nthreads > 1)
											{
												// below
												d.C0 = 0; d.C1 = 1; d.c0 = 1; d.c1 = -1;
												write_entry_A(Alocal, Blocal, texels, src, offsets, d, -1);

												d.C0 = 0; d.C1 = 1; d.c0 = 0; d.c1 = 1;
												write_entry_A(Alocal, Blocal, texels, src, offsets, d, 0);

												// above
												d.C0 = 1; d.C1 = -1; d.c0 = 1; d.c1 = -1;
												write_entry_A(Alocal, Blocal, texels, src, offsets, d, 0);

												d.C0 = 1; d.C1 = -1; d.c0 = 0; d.c1 = 1;
												write_entry_A(Alocal, Blocal, texels, src, offsets, d, 1);
											}
											else
											{
												// below
												d.C0 = 0; d.C1 = 1; d.c0 = 1; d.c1 = -1;
												write_entry_A(A, B, texels, src, offsets, d, -1);

												d.C0 = 0; d.C1 = 1; d.c0 = 0; d.c1 = 1;
												write_entry_A(A, B, texels, src, offsets, d, 0);

												// above
												d.C0 = 1; d.C1 = -1; d.c0 = 1; d.c1 = -1;
												write_entry_A(A, B, texels, src, offsets, d, 0);

												d.C0 = 1; d.C1 = -1; d.c0 = 0; d.c1 = 1;
												write_entry_A(A, B, texels, src, offsets, d, 1);
											}
#endif
											d.p10 = d.p20;
											d.p11 = d.p21;
											d.pos1 = d.pos2;
										}
									}

								}
							}

						}
					}

				}
			}

		}


		// reduce Alocal into A and Blocal into B
#pragma omp critical
		if (nthreads > 1)
		{
			A += Alocal;
			B += Blocal;
		}
	}

	// Build B
#pragma omp parallel
	{
		int nthreads = omp_get_num_threads();
		int tid = omp_get_thread_num();

		DenseTaucsMatrix Blocal;
		if (nthreads > 1)
		{
			Blocal.resize(totalsize, 3);
			Blocal.set_zero();
		}

		TrilinearClampCalcData d;
		d.avg_color = guess.img[g.num_mip_levels-1].data[0];

		for (d.i = 1; d.i < g.num_mip_levels-1; d.i++)
		{
			d.size = 1 << d.i;

			// for all texels
			for (d.y = tid; d.y < texels[0].s[1]; d.y += nthreads)
			{
				d.Oy = (d.y+d.size)%(2*d.size);
				d.Osy = (d.y+1)%2;

				for (d.x = 0; d.x < texels[0].s[0]; d.x++)
				{
					d.Ox = (d.x+d.size)%(2*d.size);
					d.Osx = (d.x+1)%2;

					Array<TexPoly> &p = texels[0](d.x, d.y).polys;

					// for all rows of matrix
					for (d.rj = 0; d.rj <= 1; d.rj++)
					{
						d.B1 = (2*d.rj - 1) / (double)(2*d.size);
						d.B0 = 1 - d.rj + d.B1*d.Oy;
						d.Ry = ((d.y+d.size)/(2*d.size) + d.rj - 1);

#ifdef MIP_CLAMP
						if (d.Ry < 0)
							d.Ry = 0;
						else if (d.Ry >= (src.s[1]/d.size))
							d.Ry = (src.s[1]/d.size) - 1;
#else
						wrap(d.Ry, src.s[1]/d.size);
#endif
						for (d.ri = 0; d.ri <= 1; d.ri++)
						{

							d.A1 = (2*d.ri - 1) / (double)(2*d.size);
							d.A0 = 1 - d.ri + d.A1*d.Ox;

							d.Rx = ((d.x+d.size)/(2*d.size) + d.ri - 1);

#ifdef MIP_CLAMP
							if (d.Rx < 0)
								d.Rx = 0;
							else if (d.Rx >= (src.s[0]/d.size))
								d.Rx = (src.s[0]/d.size) - 1;
#else
							wrap(d.Rx, src.s[0]/d.size);
#endif
							d.R = d.Ry*(src.s[0]/d.size) + d.Rx + offsets[d.i];

							// for all columns of matrix
							for (d.cj = 0; d.cj <= 1; d.cj++)
							{
								d.bs1 = (2*d.cj - 1) / 2.0;
								d.bs0 = 1 - d.cj + d.bs1*d.Osy;
								d.CBy = ((d.y+1)/2 + d.cj - 1);

#ifdef MIP_CLAMP
								if (d.CBy < 0)
									d.CBy = 0;
								else if (d.CBy >= src.s[1])
									d.CBy = src.s[1] - 1;
#else
								wrap(d.CBy, src.s[1]);
#endif
								for (d.ci = 0; d.ci <= 1; d.ci++)
								{
									d.as1 = (2*d.ci - 1) / 2.0;
									d.as0 = 1 - d.ci + d.as1*d.Osx;

									d.CBx = ((d.x+1)/2 + d.ci - 1);

#ifdef MIP_CLAMP
									if (d.CBx < 0)
										d.CBx = 0;
									else if (d.CBx >= src.s[0])
										d.CBx = src.s[0] - 1;
#else
									wrap(d.CBx, src.s[0]);
#endif

									// for all triangles
									for (int k = 0; k < p.s; k++)
									{
										d.p00 = p[k].uv[0][0];
										d.p01 = p[k].uv[0][1];
										d.pos0 = p[k].pos[0];

										d.p10 = p[k].uv[1][0];
										d.p11 = p[k].uv[1][1];
										d.pos1 = p[k].pos[1];
										for (int l = 2; l < p[k].uv.s; l++)
										{
											d.p20 = p[k].uv[l][0];
											d.p21 = p[k].uv[l][1];
											d.pos2 = p[k].pos[l];

											//** add entry to A and B
#if 0
											// bilinear
											if (nthreads > 1)
											{
												d.C0 = 1; d.C1 = 0; d.c0 = 1; d.c1 = 0;
												write_entry_B(Blocal, texels, src, offsets, d, 0);
											}
											else
											{
												d.C0 = 1; d.C1 = 0; d.c0 = 1; d.c1 = 0;
												write_entry_B(B, texels, src, offsets, d, 0);
											}
#else
											if (nthreads > 1)
											{
												// below
												d.C0 = 0; d.C1 = 1; d.c0 = 1; d.c1 = -1;
												write_entry_B(Blocal, texels, src, offsets, d, -1);

												d.C0 = 0; d.C1 = 1; d.c0 = 0; d.c1 = 1;
												write_entry_B(Blocal, texels, src, offsets, d, 0);

												// above
												d.C0 = 1; d.C1 = -1; d.c0 = 1; d.c1 = -1;
												write_entry_B(Blocal, texels, src, offsets, d, 0);

												d.C0 = 1; d.C1 = -1; d.c0 = 0; d.c1 = 1;
												write_entry_B(Blocal, texels, src, offsets, d, 1);
											}
											else
											{
												// below
												d.C0 = 0; d.C1 = 1; d.c0 = 1; d.c1 = -1;
												write_entry_B(B, texels, src, offsets, d, -1);

												d.C0 = 0; d.C1 = 1; d.c0 = 0; d.c1 = 1;
												write_entry_B(B, texels, src, offsets, d, 0);

												// above
												d.C0 = 1; d.C1 = -1; d.c0 = 1; d.c1 = -1;
												write_entry_B(B, texels, src, offsets, d, 0);

												d.C0 = 1; d.C1 = -1; d.c0 = 0; d.c1 = 1;
												write_entry_B(B, texels, src, offsets, d, 1);
											}
#endif
											d.p10 = d.p20;
											d.p11 = d.p21;
											d.pos1 = d.pos2;
										}
									}

								}
							}

						}
					}

				}
			}

		}


		// reduce Blocal into B
#pragma omp critical
		if (nthreads > 1)
		{
			B += Blocal;
		}

	}

	// a model may not use all of the texture, so constrain unused texels
	fix_unconstrained(A, B);

	//** solve
	texels.clear(); // free some memory
	A.make_ccs();
	A.data.clear();
	
	double t1 = get_time();

#ifdef SOLVE_DIRECT
	A.factorize();
	A.solve(X, B);
#elif defined(SOLVE_ITERATIVE)
	A.solve_cg(X, B);
#elif defined(SOLVE_CONSTRAINED)
	solve_constrained(A, X, B);
#endif
	double t2 = get_time();

	// free matrices
	A.clear();
	B.clear();

	//** store solution
	mip.img[0] = src;
	mip.tex[0].load(mip.img[0]);

	for (int k = 1; k < g.num_mip_levels-1; k++)
	{
		Array2D<vect4f> &dst = mip.img[k];
		
		int size = 1 << k;
		dst.resize(src.s[0]/size, src.s[1]/size);

		for (int j = 0; j < dst.s[1]; j++)
		{
			for (int i = 0; i < dst.s[0]; i++)
			{
				//printf("dst %d %d %d = %d\n", k, i, j, j*dst.s[0] + i + offsets[k]);
				dst(i,j)[0] = clamp01(X(j*dst.s[0] + i + offsets[k], 0));
				dst(i,j)[1] = clamp01(X(j*dst.s[0] + i + offsets[k], 1));
				dst(i,j)[2] = clamp01(X(j*dst.s[0] + i + offsets[k], 2));
				dst(i,j)[3] = 1;
			}
		}
	}

	mip.img[g.num_mip_levels-1] = guess.img[g.num_mip_levels-1];
	mip.tex[g.num_mip_levels-1].load(mip.img[g.num_mip_levels-1]);

	printf("--== Trilinear downscale ==--\n");
	printf("time build = %f\n", t1 - t0);
	printf("time solve = %f\n", t2 - t1);
	printf("time total = %f\n", t2 - t0);

	// upload to opengl
	upload_texture(mip);
}

