// Created: March 07, 2005
// Author: Josiah Manson
// Description:
// rand() from the C stdlib is called the number of times required to make the randomXX() bit range number.
// Thus, srand() will seed it, and it will change rand() calls and vice versa.
// ToDo: Implement seperate random number generator from the C standard one.

// Modified: March 25, 2005
// Nothing modified, just a bit more description. The problem this header addresses is that the cstdlib rand()
// function only returns 15 bits worth of random bits. Thus, 2 calls generate a 30 bit number, and if you
// want the full 32 bits, 3 calls must be made to rand().

// Modified: March 25, 2005
// Added Mersenne Twister code. I didn't like the function names, so I changed them, but it's the same otherwise.
// The website for the code is http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html

// Modified: February 11, 2010
// Added sphere code

#pragma once
#include "vect.h"

struct Random
{
	/* Period parameters */  
	static const int N = 624;
	static const int M = 397;
	static const unsigned long MATRIX_A = 0x9908b0dfUL;   /* constant vector a */
	static const unsigned long UPPER_MASK = 0x80000000UL; /* most significant w-r bits */
	static const unsigned long LOWER_MASK = 0x7fffffffUL; /* least significant r bits */

	Random()
	{
		mti=N+1;
	}

	unsigned long mt[N]; /* the array for the state vector  */
	int mti; /* mti==N+1 means mt[N] is not initialized */

	/* initializes mt[N] with a seed */
	void init(unsigned long s);

	/* initialize by an array with array-length */
	void init_array(unsigned long init_key[], int key_length);

	/* generates a random number on [0,0xffffffff]-interval */
	unsigned long int32(void);

	/* generates a random number on [0,0x7fffffff]-interval */
	long int31(void);

	/* generates a random number on [0,1]-real-interval */
	double real1(void);

	/* generates a random number on [0,1)-real-interval */
	double real2(void);

	/* generates a random number on (0,1)-real-interval */
	double real3(void);

	/* generates a random number on [0,1) with 53-bit resolution*/
	double res53(void);

	template <int N, class T>
	void in_sphere(vect<N, T> &v)
	{
		do
		{
			for (int i = 0; i < N; i++)
				v[i] = real1() * 2 - 1;
		} while (v*v > 1);
	}

	template <int N>
	vect<N, double> in_sphere()
	{
		vect<N, double> v;
		in_sphere(v);
		return v;
	}


	template <int N, class T>
	void on_sphere(vect<N, T> &v)
	{
		do
		{
			in_sphere(v);
		} while (v[0] == 0 && v[1] == 0 && v[2] == 0);

		v.normalize();
	}

	template <int N>
	vect<N, double> on_sphere()
	{
		vect<N, double> v;
		on_sphere(v);
		return v;
	}
};
