#pragma once

#include <map>
#include "array.h"
#include <assert.h>
#include "taucs.h"

using namespace std;

struct DenseTaucsMatrix;

// matrix is symmetric: 1. write vals, 2. make_css(), 3. factorize(), 4. solve.
struct SparseTaucsMatrix
{
	SparseTaucsMatrix()
	{
		A.colptr = 0;
		A.rowind = 0;
		A.values.d = 0;
		
		Aod = 0;
		L = 0;
		perm = 0;
		invperm = 0;
	}
	~SparseTaucsMatrix();

	void clear();

	Array< map<int, double> > data;
	double dummy;

	taucs_ccs_matrix A;
	taucs_ccs_matrix *Aod, *L; 
	int *perm, *invperm;

	double &operator()(int r, int c)
	{
		assert(r >= 0 && r < data.size());
		assert(c >= 0 && c < data.size());

		if (r >= c)
			return data[c][r];
		else
			return dummy;
	}

	void operator+=(SparseTaucsMatrix &b)
	{
		assert(b.data.size() == data.size());
		for (int r = 0; r < data.size(); r++)
		{
			for (map<int, double>::iterator it = b.data[r].begin(); it != b.data[r].end(); ++it)
			{
				data[r][it->first] += it->second;
			}
		}
	}

	void operator=(SparseTaucsMatrix &b)
	{
		data = b.data;
		/*clear();
		resize(b.data.s);

		for (int r = 0; r < data.size(); r++)
		{
			for (map<int, double>::iterator it = b.data[r].begin(); it != b.data[r].end(); ++it)
			{
				data[r][it->first] = it->second;
			}
		}*/
	}

	void resize(int n)
	{
		data.resize(n);
	}

	void set_zero()
	{
		dummy = 0; // why? I don't know
		for (int i = 0; i < data.size(); i++)
			data[i].clear();
	}

	void save_ccs(string fn);

	void make_ccs();
	void factorize();
	void solve(DenseTaucsMatrix &X, DenseTaucsMatrix &B);

	void solve_cg(DenseTaucsMatrix &X, DenseTaucsMatrix &B);
	
	void save_mathematica(string fn);
};

struct DenseTaucsMatrix
{
	DenseTaucsMatrix()
	{
		rows = cols = 0;
		data = 0;
	}
	
	~DenseTaucsMatrix()
	{
		clear();
	}
	void clear()
	{
		delete[] data;
		data = 0;
	}

	int rows, cols;
	double *data;
	
	void resize(int r, int c)
	{
		if (rows == r && cols == c)
			return;

		rows = r;
		cols = c;

		delete[] data;
		data = new double [rows*cols];
	}

	void set_zero()
	{
		int n = rows*cols;
		for (int i = 0; i < n; i++)
			data[i] = 0;
	}

	double &operator()(int r, int c)
	{
		assert(r >= 0 && r < rows);
		assert(c >= 0 && c < cols);

		return data[rows*c + r];
	}

	void operator+=(DenseTaucsMatrix &b)
	{
		assert(rows == b.rows && cols == b.cols);

		for (int i = 0; i < rows*cols; i++)
			data[i] += b.data[i];
	}

	
	void operator=(DenseTaucsMatrix &b)
	{
		resize(b.rows, b.cols);

		for (int i = 0; i < rows*cols; i++)
			data[i] = b.data[i];
	}

	void save_mathematica(string fn);
};

void save_dense_matrix(DenseTaucsMatrix &x, string fn);
void load_dense_matrix(DenseTaucsMatrix &x, string fn);
